# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from builtins import object
try:
    # Python 3
    from unittest.mock import Mock
except:
    # Python 2 (library)
    from mock import Mock
import os
import twisted
from twisted.internet import defer
from twisted.python.versions import Version
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.address import I2PAddress
from txi2p.sam import session
from txi2p.sam.constants import DEFAULT_SIGTYPE
from txi2p.test.util import TEST_B64
from .util import SAMProtocolTestMixin, SAMFactoryTestMixin

if twisted.version < Version('twisted', 12, 3, 0):
    skipSRO = 'TestCase.successResultOf() requires twisted 12.3 or newer'
else:
    skipSRO = None


class TestSessionCreateProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = session.SessionCreateProtocol

    def test_sessionCreateAfterHello(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=%s\n' % DEFAULT_SIGTYPE,
            proto.transport.value().decode('utf-8'))

    def test_sessionCreateAfterHelloWith3point0(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.0\n')
        self.assertEquals(
            b'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT\n',
            proto.transport.value())

    def test_sessionCreateAfterHelloWithSigType(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.sigType = 'foobar'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=foobar\n',
            proto.transport.value())

    def test_sessionCreateWithAutoNickAfterHello(self):
        fac, proto = self.makeProto()
        fac.nickname = None
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            'SESSION CREATE STYLE=STREAM ID=txi2p-%s DESTINATION=TRANSIENT SIGNATURE_TYPE=%s\n' % (os.getpid(), DEFAULT_SIGTYPE),
            proto.transport.value().decode('utf-8'))

    def test_sessionCreateWithPortAfterHello(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.localPort = 81
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=%s FROM_PORT=81\n' % DEFAULT_SIGTYPE,
            proto.transport.value().decode('utf-8'))

    def test_sessionCreateWithOptionsAfterHello(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.options['bar'] = 'baz'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=%s bar=baz\n' % DEFAULT_SIGTYPE,
            proto.transport.value().decode('utf-8'))

    def test_sessionCreateFallbackForUnsupportedDest(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'SESSION STATUS RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE foobar unsupported"\n')
        self.assertEquals(
            b'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=ECDSA_SHA256_P256\n',
            proto.transport.value())

    def test_sessionCreateSecondFallbackForUnsupportedDest(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'SESSION STATUS RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE ECDSA_SHA256_P256 unsupported"\n')
        self.assertEquals(
            b'SESSION CREATE STYLE=STREAM ID=foo DESTINATION=TRANSIENT SIGNATURE_TYPE=DSA_SHA1\n',
            proto.transport.value())

    def test_sessionCreateReturnsUnsupportedDestWithSigType(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.sigType = DEFAULT_SIGTYPE
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(('SESSION STATUS RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE %s unsupported"\n' % DEFAULT_SIGTYPE).encode('utf-8'))
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'SIGNATURE_TYPE %s unsupported' % DEFAULT_SIGTYPE)

    def test_sessionCreateReturnsUnsupportedDestAgainWithSigType(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.sigType = 'ECDSA_SHA256_P256'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'SESSION STATUS RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE ECDSA_SHA256_P256 unsupported"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'SIGNATURE_TYPE ECDSA_SHA256_P256 unsupported')

    def test_sessionCreateReturnsError(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'SESSION STATUS RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_namingLookupAfterSessionCreate(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(('SESSION STATUS RESULT=OK DESTINATION=%s\n' % TEST_B64).encode('utf-8'))
        self.assertEquals(
            b'NAMING LOOKUP NAME=ME\n',
            proto.transport.value())

    def test_sessionCreatedAfterNamingLookup(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.sessionCreated = Mock()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(('SESSION STATUS RESULT=OK DESTINATION=%s\n' % TEST_B64).encode('utf-8'))
        proto.transport.clear()
        proto.dataReceived(('NAMING REPLY RESULT=OK NAME=ME VALUE=%s\n' % TEST_B64).encode('utf-8'))
        fac.sessionCreated.assert_called_with(proto.receiver, TEST_B64)

    def test_sessionCreatedAndKeepaliveStartedAfterNamingLookupWith3_2(self):
        fac, proto = self.makeProto()
        fac.style = 'STREAM'
        fac.sessionCreated = Mock()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.2\n')
        proto.transport.clear()
        proto.dataReceived(('SESSION STATUS RESULT=OK DESTINATION=%s\n' % TEST_B64).encode('utf-8'))
        proto.transport.clear()
        proto.dataReceived(('NAMING REPLY RESULT=OK NAME=ME VALUE=%s\n' % TEST_B64).encode('utf-8'))
        self.assertEquals('State_keepalive', proto.receiver.currentRule)
        fac.sessionCreated.assert_called_with(proto.receiver, TEST_B64)
        # Cleanup
        self.addCleanup(proto.receiver.stopPinging)


class FakeEndpoint(object):
    def __init__(self, failure=None):
        self.failure = failure
        self.deferred = None
        self.facDeferred = None
        self.called = 0

    def connect(self, fac):
        self.called += 1
        fac.deferred = self.facDeferred
        self.factory = fac
        return self.deferred


class TestSessionCreateFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = session.SessionCreateFactory
    blankFactoryArgs = ('',)

    def test_startFactory(self):
        tmp = '/tmp/TestSessionCreateFactory.privKey'
        fac, proto = self.makeProto('foo', keyfile=tmp)
        fac.doStart()
        self.assertTrue(fac._writeKeypair)

    def test_startFactoryWithExistingKeyfile(self):
        tmp = '/tmp/TestSessionCreateFactory.privKey'
        f = open(tmp, 'w')
        f.write(u'foo')
        f.close()
        fac, proto = self.makeProto('foo', keyfile=tmp)
        fac.doStart()
        self.assertEqual('foo', fac.privKey)
        os.remove(tmp)

    def test_sessionCreated(self):
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto('foo')
        fac.samVersion = '3.1'
        # Shortcut to end of SAM session create protocol
        proto.receiver.currentRule = 'State_naming'
        proto._parser._setupInterp()
        proto.dataReceived(('NAMING REPLY RESULT=OK NAME=ME VALUE=%s\n' % TEST_B64).encode('utf-8'))
        s = self.successResultOf(fac.deferred)
        self.assertEqual(('3.1', 'STREAM', 'foo', proto.receiver, TEST_B64, None), s)
    test_sessionCreated.skip = skipSRO

    def test_sessionCreatedWithKeyfile(self):
        tmp = '/tmp/TestSessionCreateFactory.privKey'
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto('foo', keyfile=tmp)
        fac.samVersion = '3.1'
        fac.privKey = 'bar'
        fac._writeKeypair = True
        # Shortcut to end of SAM session create protocol
        proto.receiver.currentRule = 'State_naming'
        proto._parser._setupInterp()
        proto.dataReceived(('NAMING REPLY RESULT=OK NAME=ME VALUE=%s\n' % TEST_B64).encode('utf-8'))
        f = open(tmp, 'r')
        privKey = f.read()
        f.close()
        self.assertEqual('bar', privKey)
        # Cleanup
        os.remove(tmp)


class TestSAMSession(unittest.TestCase):
    def setUp(self):
        self.tr = proto_helpers.StringTransportWithDisconnection()
        proto = Mock()
        proto.sender = Mock()
        proto.sender.transport = self.tr
        self.tr.protocol = proto
        self.s = session.SAMSession()
        self.s.nickname = 'foo'
        self.s.samVersion = '3.1'
        self.s.id = 'foo'
        self.s._proto = proto
        session._sessions['foo'] = self.s

    def tearDown(self):
        session._sessions = {}

    def test_addStream(self):
        self.assertEqual([], self.s._streams)
        self.s.addStream('foo')
        self.assertEqual(['foo'], self.s._streams)

    def test_removeStream_autoClose(self):
        self.s._autoClose = True
        self.s.addStream('bar')
        self.s.addStream('baz')
        self.s.removeStream('bar')
        self.assertEqual(['baz'], self.s._streams)
        self.assertEqual(True, 'foo' in session._sessions)
        self.assertEqual(self.s, session._sessions['foo'])
        self.s.removeStream('baz')
        self.assertEqual([], self.s._streams)
        self.assertEqual({}, session._sessions)

    def test_removeStream_noAutoClose(self):
        self.s.addStream('bar')
        self.s.addStream('baz')
        self.s.removeStream('bar')
        self.assertEqual(['baz'], self.s._streams)
        self.assertEqual(True, 'foo' in session._sessions)
        self.assertEqual(self.s, session._sessions['foo'])
        self.s.removeStream('baz')
        self.assertEqual([], self.s._streams)
        self.assertEqual({'foo': self.s}, session._sessions)


class TestGetSession(unittest.TestCase):
    def tearDown(self):
        session._sessions = {}

    def test_getSession_newNickname(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(('3.1', 'STREAM', 'nick', proto, TEST_B64, None))
        d = session.getSession('nick', samEndpoint)
        s = self.successResultOf(d)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual('nick', s.nickname)
        self.assertEqual('nick', s.id)
        self.assertEqual(proto, s._proto)
        self.assertEqual(TEST_B64, s.address.destination)
        self.assertEqual(None, s.address.port)
    test_getSession_newNickname.skip = skipSRO

    def test_getSession_newNickname_withPort(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(('3.2', 'STREAM', 'nick', proto, TEST_B64, 81))
        d = session.getSession('nick', samEndpoint)
        s = self.successResultOf(d)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual('nick', s.nickname)
        self.assertEqual('nick', s.id)
        self.assertEqual(proto, s._proto)
        self.assertEqual(TEST_B64, s.address.destination)
        self.assertEqual(81, s.address.port)
    test_getSession_newNickname_withPort.skip = skipSRO

    def test_getSession_newNickname_withoutEndpoint(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(('3.1', 'STREAM', 'nick', proto, TEST_B64, None))
        self.assertRaises(ValueError, session.getSession, 'nick')
    test_getSession_newNickname_withoutEndpoint.skip = skipSRO

    def test_getSession_existingNickname(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(('3.1', 'STREAM', 'nick', proto, TEST_B64, None))
        d = session.getSession('nick', samEndpoint)
        s = self.successResultOf(d)
        d2 = session.getSession('nick', samEndpoint)
        s2 = self.successResultOf(d2)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual(s, s2)
    test_getSession_existingNickname.skip = skipSRO

    def test_getSession_existingNickname_withoutEndpoint(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(('3.1', 'STREAM', 'nick', proto, TEST_B64, None))
        d = session.getSession('nick', samEndpoint)
        s = self.successResultOf(d)
        d2 = session.getSession('nick')
        s2 = self.successResultOf(d2)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual(s, s2)
    test_getSession_existingNickname_withoutEndpoint.skip = skipSRO


class TestDestGenerateProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = session.DestGenerateProtocol

    def test_destGenerateAfterHello(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals('DEST GENERATE SIGNATURE_TYPE=%s\n' % DEFAULT_SIGTYPE, proto.transport.value().decode('utf-8'))

    def test_destGenerateAfterHelloWith3point0(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.0\n')
        self.assertEquals(b'DEST GENERATE\n', proto.transport.value())

    def test_destGenerateAfterHelloWithSigType(self):
        fac, proto = self.makeProto()
        fac.sigType = 'foobar'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(b'DEST GENERATE SIGNATURE_TYPE=foobar\n', proto.transport.value())

    def test_destGeneratedFallbackForUnsupportedDest(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(('DEST REPLY RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE %s unsupported"\n' % DEFAULT_SIGTYPE).encode('utf-8'))
        self.assertEquals(b'DEST GENERATE SIGNATURE_TYPE=ECDSA_SHA256_P256\n', proto.transport.value())

    def test_destGeneratedSecondFallbackForUnsupportedDest(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'DEST REPLY RESULT=I2P_ERROR MESSAGE="SIGNATURE_TYPE ECDSA_SHA256_P256 unsupported"\n')
        self.assertEquals(b'DEST GENERATE SIGNATURE_TYPE=DSA_SHA1\n', proto.transport.value())

    def test_destGeneratedReturnsError(self):
        fac, proto = self.makeProto()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'DEST REPLY RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_destGenerated(self):
        fac, proto = self.makeProto()
        fac.destGenerated = Mock()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(('DEST REPLY PUB=%s PRIV=%s\n' % (TEST_B64, 'TEST_PRIV')).encode('utf-8'))
        fac.destGenerated.assert_called_with(TEST_B64, 'TEST_PRIV')


class TestDestGenerateFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = session.DestGenerateFactory
    blankFactoryArgs = ('', None)

    def test_destGenerated(self):
        tmp = '/tmp/TestDestGenerateFactory.privKey'
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto(tmp)
        # Shortcut to end of SAM dest generate protocol
        proto.receiver.currentRule = 'State_dest'
        proto._parser._setupInterp()
        proto.dataReceived(('DEST REPLY PUB=%s PRIV=%s\n' % (TEST_B64, 'TEST_PRIV')).encode('utf-8'))
        s = self.successResultOf(fac.deferred)
        self.assertEqual(I2PAddress(TEST_B64), s)
        os.remove(tmp)
    test_destGenerated.skip = skipSRO

    def test_destGenerated_privKeySaved(self):
        tmp = '/tmp/TestDestGenerateFactory.privKey'
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto(tmp)
        # Shortcut to end of SAM dest generate protocol
        proto.receiver.currentRule = 'State_dest'
        proto._parser._setupInterp()
        proto.dataReceived(('DEST REPLY PUB=%s PRIV=%s\n' % (TEST_B64, 'TEST_PRIV')).encode('utf-8'))
        f = open(tmp, 'r')
        privKey = f.read()
        f.close()
        self.assertEqual('TEST_PRIV', privKey)
        os.remove(tmp)

    def test_destGenerated_keyfileExists(self):
        tmp = '/tmp/TestDestGenerateFactory.privKey'
        f = open(tmp, 'w')
        f.write(u'foo')
        f.close()
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto(tmp)
        # Shortcut to end of SAM dest generate protocol
        proto.receiver.currentRule = 'State_dest'
        proto._parser._setupInterp()
        proto.dataReceived(('DEST REPLY PUB=%s PRIV=%s\n' % (TEST_B64, 'TEST_PRIV')).encode('utf-8'))
        self.assertIsInstance(self.failureResultOf(fac.deferred).value, ValueError)
        os.remove(tmp)
    test_destGenerated_keyfileExists.skip = skipSRO


class TestGenerateDestination(unittest.TestCase):
    def test_generateDestination(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(I2PAddress(TEST_B64))
        d = session.generateDestination('', samEndpoint)
        s = self.successResultOf(d)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual(I2PAddress(TEST_B64), s)
    test_generateDestination.skip = skipSRO


class TestTestAPIProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = session.TestAPIProtocol

    def test_samConnected(self):
        fac, proto = self.makeProto()
        fac.samConnected = Mock()
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        fac.samConnected.assert_called_with()


class TestTestAPIFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = session.TestAPIFactory
    blankFactoryArgs = []

    def test_samConnected(self):
        mreactor = proto_helpers.MemoryReactor()
        fac, proto = self.makeProto()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        s = self.successResultOf(fac.deferred)
        self.assertEqual(True, s)
    test_samConnected.skip = skipSRO


class TestTestAPI(unittest.TestCase):
    def test_testAPI(self):
        proto = proto_helpers.AccumulatingProtocol()
        samEndpoint = FakeEndpoint()
        samEndpoint.deferred = defer.succeed(None)
        samEndpoint.facDeferred = defer.succeed(True)
        d = session.testAPI(samEndpoint)
        s = self.successResultOf(d)
        self.assertEqual(1, samEndpoint.called)
        self.assertEqual(True, s)
    test_testAPI.skip = skipSRO



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.
from __future__ import print_function

from builtins import object
import os
import sys
from twisted.internet import defer, error
from twisted.python import failure, log

from txi2p import grammar
from txi2p.address import I2PAddress
from txi2p.sam import constants as c
from txi2p.sam.base import (
    cmpSAM,
    makeSAMProtocol,
    SAMSender,
    SAMReceiver,
    SAMFactory,
)

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class SessionCreateSender(SAMSender):
    def sendSessionCreate(self, samVersion, style, id, privKey, localPort, options, sigType):
        msg = 'SESSION CREATE'
        msg += ' STYLE=%s' % style
        msg += ' ID=%s' % id
        msg += ' DESTINATION=%s' % (privKey if privKey else 'TRANSIENT')
        if cmpSAM(samVersion, '3.1') >= 0 and not privKey:
            msg += ' SIGNATURE_TYPE=%s' % (sigType and sigType or c.DEFAULT_SIGTYPE)
        if localPort:
            msg += ' FROM_PORT=%d' % localPort
        for key in options:
            msg += ' %s=%s' % (key, options[key])
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))


class SessionCreateReceiver(SAMReceiver):
    def command(self):
        if not (hasattr(self.factory, 'nickname') and self.factory.nickname):
            # All tunnels in the same process use the same nickname
            # TODO is using the PID a security risk?
            self.factory.nickname = 'txi2p-%d' % os.getpid()

        self.sender.sendSessionCreate(
            self.factory.samVersion,
            self.factory.style,
            self.factory.nickname,
            self.factory.privKey,
            self.factory.localPort,
            self.factory.options,
            self.factory.sigType)
        self.currentRule = 'State_create'

    def create(self, result, destination=None, message=None):
        if result != c.RESULT_OK:
            # If the user didn't specify a SigType, try falling back
            if cmpSAM(self.factory.samVersion, '3.1') >= 0 and \
                    message.startswith('SIGNATURE_TYPE') and \
                    not self.factory.sigType:
                fallback = 'ECDSA_SHA256_P256' in message and 'DSA_SHA1' or 'ECDSA_SHA256_P256'
                eprint('Warning: %s, falling back to %s' % (message, fallback))
                self.sender.sendSessionCreate(
                    self.factory.samVersion,
                    self.factory.style,
                    self.factory.nickname,
                    self.factory.privKey,
                    self.factory.localPort,
                    self.factory.options,
                    fallback)
            else:
                self.factory.resultNotOK(result, message)
            return

        self.factory.privKey = destination
        self.sender.sendNamingLookup('ME')
        self.currentRule = 'State_naming'

    def postLookup(self, dest):
        # Help keep the session open
        if cmpSAM(self.factory.samVersion, '3.2') >= 0:
            self.startPinging()
        else:
            try:
                self.sender.transport.setTcpKeepAlive(1)
            except AttributeError as e:
                eprint(e)
        self.factory.sessionCreated(self, dest)


# A Protocol for making a SAM session
SessionCreateProtocol = makeSAMProtocol(
    SessionCreateSender,
    SessionCreateReceiver)


class SessionCreateFactory(SAMFactory):
    protocol = SessionCreateProtocol

    def __init__(self, nickname, style='STREAM', keyfile=None, localPort=None, options=None, sigType=None):
        if style != 'STREAM':
            raise error.UnsupportedSocketType()
        if options is None:
            options = {}
        self.nickname = nickname
        self.style = style
        self._keyfile = keyfile
        self.localPort = localPort
        self.options = options
        self.sigType = sigType
        self.deferred = defer.Deferred(self._cancel)
        self.samVersion = None
        self.privKey = None
        self._writeKeypair = False

    def startFactory(self):
        if self._keyfile:
            try:
                f = open(self._keyfile, 'r')
                self.privKey = f.read()
                f.close()
            except IOError:
                log.msg('Could not load private key from %s' % self._keyfile)
                self._writeKeypair = True

    def sessionCreated(self, proto, pubKey):
        if self._writeKeypair:
            try:
                f = open(self._keyfile, 'w')
                f.write(str(self.privKey))
                f.close()
            except IOError:
                log.msg('Could not save private key to %s' % self._keyfile)
        # Now continue on with creation of SAMSession
        self.deferred.callback((self.samVersion, self.style, self.nickname, proto, pubKey, self.localPort))


# Dictionary containing all active SAM sessions
_sessions = {}
# Dictionary containing all pending SAM sessions
_pending_sessions = {}


class SAMSession(object):
    """A SAM session represents an active I2P Destination.

    Attributes:
        nickname (str): The user-assigned session nickname, can be `None`.
        samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that will connect to the SAM API.
        samVersion (str): The SAM version in use by this session.
        style (str): The session style.
        id (str): SAM Session ID, autogenerated if ``nickname`` is None, else
            ``nickname``.
        address (txi2p.I2PAddress): The Destination of this session.
    """

    def __init__(self):
        self.nickname = None
        self.samEndpoint = None
        self.samVersion = ''
        self.style = 'STREAM'
        self.id = None
        self.address = None
        self._proto = None
        self._autoClose = False
        self._closed = False
        self._streams = []

    def addStream(self, stream):
        """Register a stream with this session.

        Raises:
            twisted.internet.error.ConnectionDone: if the session is closed.
        """
        if self._closed:
            raise error.ConnectionDone
        self._streams.append(stream)

    def removeStream(self, stream):
        """Remove a stream from this session.

        If this was the last stream, and the session was set to auto-close, the
        session will close and become unusable.

        Raises:
            twisted.internet.error.ConnectionDone: if the session is closed.
        """
        if self._closed:
            raise error.ConnectionDone
        # Streams are only added once they have been established
        if stream in self._streams:
            self._streams.remove(stream)
        if not self._streams and self._autoClose:
            # No more streams, close the session
            self.close()

    def close(self):
        """Close the session."""
        self._closed = True
        self._streams = []
        self._proto.sender.transport.loseConnection()
        del _sessions[self.nickname]


def getSession(nickname, samEndpoint=None, autoClose=False, **kwargs):
    """Get or create a SAM session.

    Args:
        nickname (str): The session nickname.
        samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that will connect to the SAM API.
        autoClose (bool): `true` if the session should close automatically once
            no more connections are using it.
    """
    if nickname in _sessions:
        return defer.succeed(_sessions[nickname])
    elif nickname in _pending_sessions:
        def cancel(d):
            if nickname in _pending_sessions and d in _pending_sessions[nickname]:
                _pending_sessions[nickname].remove(d)
        d = defer.Deferred(cancel)
        _pending_sessions[nickname].append(d)
        return d

    if not samEndpoint:
        raise ValueError('A new session cannot be created without an API Endpoint')

    def createSession(xxx_todo_changeme):
        (samVersion, style, id, proto, pubKey, localPort) = xxx_todo_changeme
        s = SAMSession()
        s.nickname = nickname
        s.samEndpoint = samEndpoint
        s.samVersion = samVersion
        s.style = style
        s.id = id
        s.address = I2PAddress(pubKey, port=localPort)
        s._proto = proto
        s._autoClose = autoClose
        _sessions[nickname] = s

        waiting = _pending_sessions.pop(nickname, [])
        for d in waiting:
            d.callback(s)

        return s

    def errbackPending(f):
        waiting = _pending_sessions.pop(nickname, [])
        for d in waiting:
            d.errback(f)
        return f

    _pending_sessions[nickname] = []
    sessionFac = SessionCreateFactory(nickname, **kwargs)
    d = samEndpoint.connect(sessionFac)
    # Force caller to wait until the session is actually created
    d.addCallback(lambda proto: sessionFac.deferred)
    d.addCallback(createSession)
    d.addErrback(errbackPending)
    return d


class DestGenerateSender(SAMSender):
    def sendDestGenerate(self, samVersion, sigType=None):
        msg = 'DEST GENERATE'
        if cmpSAM(samVersion, '3.1') >= 0:
            msg += ' SIGNATURE_TYPE=%s' % (sigType and sigType or 'EdDSA_SHA512_Ed25519')
        msg += '\n'
        self.transport.write(msg.encode('utf-8'))


class DestGenerateReceiver(SAMReceiver):
    def command(self):
        self.sender.sendDestGenerate(
            self.factory.samVersion,
            self.factory.sigType)
        self.currentRule = 'State_dest'

    def destGenerated(self, result=None, pub=None, priv=None, message=None):
        if result:
            # If the user didn't specify a SigType, try falling back
            if cmpSAM(self.factory.samVersion, '3.1') >= 0 and \
                    message.startswith('SIGNATURE_TYPE') and \
                    not self.factory.sigType:
                fallback = 'ECDSA_SHA256_P256' in message and 'DSA_SHA1' or 'ECDSA_SHA256_P256'
                eprint('Warning: %s, falling back to %s' % (message, fallback))
                self.sender.sendDestGenerate(
                    self.factory.samVersion,
                    fallback)
            else:
                self.factory.resultNotOK(result, message)
            return

        self.factory.destGenerated(pub, priv)
        self.sender.transport.loseConnection()


# A Protocol for generating an I2P Destination via SAM
DestGenerateProtocol = makeSAMProtocol(
    DestGenerateSender,
    DestGenerateReceiver)


class DestGenerateFactory(SAMFactory):
    protocol = DestGenerateProtocol

    def __init__(self, keyfile, sigType=None):
        self._keyfile = keyfile
        self.sigType = sigType
        self.deferred = defer.Deferred(self._cancel)

    def destGenerated(self, pubKey, privKey):
        if os.path.exists(self._keyfile):
            self.deferred.errback(failure.Failure(ValueError('The keyfile already exists')))
            return

        try:
            f = open(self._keyfile, 'w')
            f.write(str(privKey))
            f.close()
            self.deferred.callback(I2PAddress(pubKey))
        except IOError as e:
            self.deferred.errback(failure.Failure(e))


def generateDestination(keyfile, samEndpoint, sigType=None):
    """Generate a new I2P Destination.

    The function returns a :class:`twisted.internet.defer.Deferred`; register
    callbacks to receive the return value or errors.

    Args:
        keyfile (str): Path to a local file where the keypair for the new
            Destination should be stored.
        samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that will connect to the SAM API.
        sigType (str): The SigType to generate. Defaults to Ed25519 if
            supported, falling back to ECDSA_SHA256_P256 and then DSA_SHA1.

    Returns:
        txi2p.I2PAddress: The new Destination. Once this is received via the
        Deferred callback, the ``keyfile`` will have been written.

    Raises:
        ValueError: if the ``keyfile`` already exists.
        IOError: if the ``keyfile`` write fails.
    """
    destFac = DestGenerateFactory(keyfile, sigType)
    d = samEndpoint.connect(destFac)
    d.addCallback(lambda proto: destFac.deferred)
    return d


class TestAPIReceiver(SAMReceiver):
    def command(self):
        self.factory.samConnected()
        self.sender.transport.loseConnection()


# A Protocol for testing whether a SAM API is reachable
TestAPIProtocol = makeSAMProtocol(
    SAMSender,
    TestAPIReceiver)


class TestAPIFactory(SAMFactory):
    protocol = TestAPIProtocol

    def __init__(self):
        self.deferred = defer.Deferred(self._cancel)

    def samConnected(self):
        self.deferred.callback(True)


def testAPI(samEndpoint):
    """Test whether a SAM API is reachable.

    The function returns a :class:`twisted.internet.defer.Deferred`; register
    callbacks to receive the return value or errors.

    Args:
        samEndpoint (twisted.internet.interfaces.IStreamClientEndpoint): An
            endpoint that may connect to the SAM API.

    Returns:
        True if the API is reachable.
    """
    testFac = TestAPIFactory()
    d = samEndpoint.connect(testFac)
    d.addCallback(lambda proto: testFac.deferred)
    return d



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

try:
    # Python 3
    from unittest.mock import Mock
except:
    # Python 2 (library)
    from mock import Mock
import os
import twisted
from twisted.internet import defer
from twisted.python.versions import Version
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.address import I2PAddress
from txi2p.sam import stream
from txi2p.test.util import TEST_B64, FakeFactory
from .util import SAMProtocolTestMixin, SAMFactoryTestMixin

if twisted.version < Version('twisted', 12, 3, 0):
    skipSRO = 'TestCase.successResultOf() requires twisted 12.3 or newer'
else:
    skipSRO = None


class TestStreamConnectProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = stream.StreamConnectProtocol

    def test_streamConnectAfterHello(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = 'bar'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'STREAM CONNECT ID=foo DESTINATION=bar SILENT=false\n',
            proto.transport.value())

    def test_streamConnectWithPortAfterHello(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = 'bar'
        fac.port = 80
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'STREAM CONNECT ID=foo DESTINATION=bar SILENT=false TO_PORT=80\n',
            proto.transport.value())

    def test_streamConnectWithLocalPortAfterHello(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = 'bar'
        fac.localPort = 34444
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'STREAM CONNECT ID=foo DESTINATION=bar SILENT=false FROM_PORT=34444\n',
            proto.transport.value())

    def test_namingLookupAfterHello(self):
        fac, proto = self.makeProto()
        fac.dest = None
        fac.host = 'spam.i2p'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'NAMING LOOKUP NAME=spam.i2p\n',
            proto.transport.value())

    def test_namingLookupReturnsError(self):
        fac, proto = self.makeProto()
        fac.dest = None
        fac.host = 'spam.i2p'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'NAMING REPLY RESULT=KEY_NOT_FOUND NAME=spam.i2p MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('KEY_NOT_FOUND', 'foo bar baz')

    def test_streamConnectAfterNamingLookup(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = None
        fac.host = 'spam.i2p'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'NAMING REPLY RESULT=OK NAME=spam.i2p VALUE=bar\n')
        self.assertEquals(
            b'STREAM CONNECT ID=foo DESTINATION=bar SILENT=false\n',
            proto.transport.value())

    def test_streamConnectReturnsError(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = 'bar'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_streamConnectionEstablishedAfterStreamConnect(self):
        fac, proto = self.makeProto()
        fac.streamConnectionEstablished = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.dest = 'bar'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        fac.streamConnectionEstablished.assert_called_with(proto.receiver)


class TestStreamConnectFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = stream.StreamConnectFactory
    blankFactoryArgs = (None, Mock(), '', '')

    def test_streamConnectionEstablished(self):
        mreactor = proto_helpers.MemoryReactor()
        wrappedFactory = FakeFactory()
        session = Mock()
        fac, proto = self.makeProto(wrappedFactory, session, 'spam.i2p', 'foo')
        # Shortcut to end of SAM stream connect protocol
        proto.receiver.currentRule = 'State_connect'
        proto._parser._setupInterp()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        session.addStream.assert_called_with(proto.receiver)
        streamProto = self.successResultOf(fac.deferred)
        self.assertEqual(proto.receiver.wrappedProto, streamProto)
    test_streamConnectionEstablished.skip = skipSRO


class TestStreamAcceptProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = stream.StreamAcceptProtocol

    def test_streamAcceptAfterHello(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'STREAM ACCEPT ID=foo SILENT=false\n',
            proto.transport.value())

    def test_streamAcceptReturnsError(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_streamAcceptEstablishedAfterStreamAccept(self):
        fac, proto = self.makeProto()
        fac.streamAcceptEstablished = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        fac.streamAcceptEstablished.assert_called_with(proto.receiver)

    def test_streamAcceptIncomingAfterPeerAddress(self):
        fac, proto = self.makeProto()
        fac.streamAcceptIncoming = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_readData'
        proto._parser._setupInterp()
        proto.dataReceived(('%s FROM_PORT=34444 TO_PORT=0\n' % TEST_B64).encode('utf-8'))
        fac.streamAcceptIncoming.assert_called_with(proto.receiver)
        self.assertEquals(
            I2PAddress(TEST_B64, port=34444),
            proto.receiver.peer)

    def test_peerDataWrapped_allAtOnce(self):
        fac, proto = self.makeProto()
        fac.streamAcceptIncoming = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.receiver.wrappedProto = proto_helpers.AccumulatingProtocol()
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_readData'
        proto._parser._setupInterp()
        proto.dataReceived(('%s FROM_PORT=34444 TO_PORT=0\nEgg and spam' % TEST_B64).encode('utf-8'))
        self.assertEquals(b'Egg and spam', proto.receiver.wrappedProto.data)

    def test_peerDataWrapped_twoParts(self):
        fac, proto = self.makeProto()
        fac.streamAcceptIncoming = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.receiver.wrappedProto = proto_helpers.AccumulatingProtocol()
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_readData'
        proto._parser._setupInterp()
        proto.dataReceived(('%s FROM_PORT=34444 TO_PORT=0\nEgg a' % TEST_B64).encode('utf-8'))
        self.assertEquals(b'Egg a', proto.receiver.wrappedProto.data)
        proto.dataReceived(b'nd spam')
        self.assertEquals(b'Egg and spam', proto.receiver.wrappedProto.data)

    def test_peerDataWrapped_threeParts(self):
        fac, proto = self.makeProto()
        fac.streamAcceptIncoming = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.receiver.wrappedProto = proto_helpers.AccumulatingProtocol()
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_readData'
        proto._parser._setupInterp()
        proto.dataReceived(('%s FROM_PORT=34444 T' % TEST_B64).encode('utf-8'))
        proto.dataReceived(b'O_PORT=0\nEgg a')
        self.assertEquals(b'Egg a', proto.receiver.wrappedProto.data)
        proto.dataReceived(b'nd spam')
        self.assertEquals(b'Egg and spam', proto.receiver.wrappedProto.data)


class TestStreamAcceptFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = stream.StreamAcceptFactory
    blankFactoryArgs = (Mock(), Mock(), Mock())

    def test_streamAcceptEstablished(self):
        mreactor = proto_helpers.MemoryReactor()
        session = Mock()
        listeningPort = Mock()
        fac, proto = self.makeProto(Mock(), session, listeningPort)
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_accept'
        proto._parser._setupInterp()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        session.addStream.assert_called_with(proto.receiver)
        listeningPort.addAccept.assert_called_with(proto.receiver)
    test_streamAcceptEstablished.skip = skipSRO

    def test_streamAcceptIncoming(self):
        mreactor = proto_helpers.MemoryReactor()
        wrappedFactory = FakeFactory()
        session = Mock()
        listeningPort = Mock()
        fac, proto = self.makeProto(wrappedFactory, session, listeningPort)
        # Shortcut to end of SAM stream accept protocol
        proto.receiver.currentRule = 'State_readData'
        proto._parser._setupInterp()
        proto.dataReceived(('%s FROM_PORT=34444 TO_PORT=0\n' % TEST_B64).encode('utf-8'))
        listeningPort.removeAccept.assert_called_with(proto.receiver)
        self.assertEqual(wrappedFactory.proto, proto.receiver.wrappedProto)
    test_streamAcceptEstablished.skip = skipSRO


class TestStreamForwardProtocol(SAMProtocolTestMixin, unittest.TestCase):
    protocol = stream.StreamForwardProtocol

    def test_streamForwardAfterHello(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        self.assertEquals(
            b'STREAM FORWARD ID=foo PORT=1337 SILENT=false\n',
            proto.transport.value())

    def test_streamForwardReturnsError(self):
        fac, proto = self.makeProto()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=I2P_ERROR MESSAGE="foo bar baz"\n')
        fac.resultNotOK.assert_called_with('I2P_ERROR', 'foo bar baz')

    def test_streamForwardEstablishedAfterStreamForward(self):
        fac, proto = self.makeProto()
        fac.streamForwardEstablished = Mock()
        fac.session = Mock()
        fac.session.id = 'foo'
        fac.forwardPort = 1337
        proto.transport.clear()
        proto.dataReceived(b'HELLO REPLY RESULT=OK VERSION=3.1\n')
        proto.transport.clear()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        fac.streamForwardEstablished.assert_called_with(proto.receiver)


class TestStreamForwardFactory(SAMFactoryTestMixin, unittest.TestCase):
    factory = stream.StreamForwardFactory
    blankFactoryArgs = (Mock(), '')

    def test_streamForwardEstablished(self):
        mreactor = proto_helpers.MemoryReactor()
        session = Mock()
        fac, proto = self.makeProto(session, '')
        # Shortcut to end of SAM stream forward protocol
        proto.receiver.currentRule = 'State_forward'
        proto._parser._setupInterp()
        proto.dataReceived(b'STREAM STATUS RESULT=OK\n')
        session.addStream.assert_called_with(proto.receiver)
        streamProto = self.successResultOf(fac.deferred)
        self.assertEqual(proto.receiver, streamProto)
    test_streamForwardEstablished.skip = skipSRO
